DUMMY_PARENT_NAME = "parent"
DUMMY_CHILD_NAME = "child"
PARENT_SHORT_NAME = "p"
CHILD_SHORT_NAME = "c"
DUMMY_ENV_PREFIX = "GW"
DUMMY_ATN_NAME = "dummy_atn1"
DUMMY_SCADA1_NAME = "dummy_scada1"
DUMMY_SCADA2_NAME = "dummy_scada2"
DUMMY_ADMIN_NAME = "dummy_admin"
DUMMY_ATN_SHORT_NAME = "a"
DUMMY_SCADA1_SHORT_NAME = "s"
DUMMY_SCADA2_SHORT_NAME = "s2"
DUMMY_ADMIN_SHORT_NAME = "dummy_admin"
DUMMY_CHILD_ENV_PREFIX = DUMMY_ENV_PREFIX + "CHILD_"
DUMMY_PARENT_ENV_PREFIX = DUMMY_ENV_PREFIX + "PARENT_"
DUMMY_ATN_ENV_PREFIX = DUMMY_ENV_PREFIX + "ATN1_"
DUMMY_SCADA1_ENV_PREFIX = DUMMY_ENV_PREFIX + "SCADA1_"
DUMMY_SCADA2_ENV_PREFIX = DUMMY_ENV_PREFIX + "SCADA1_"
